import concurrent
import math
import os
import subprocess
import sys
import time
import warnings
from argparse import Namespace

# subprocess.check_call([sys.executable, "-m", "pip", "install", "cupy"])
from pathlib import Path

import cv2
import h5py
import numpy as np
import torchvision
from flow_vis import flow_vis
from scipy.io import loadmat
from torch.autograd import Variable

import blur2flowNet

warnings.filterwarnings("ignore", category=FutureWarning)

import numpy
import torch
from torch.utils.data import DataLoader
from torchvision import transforms

torch.backends.cudnn.enabled = True  # make sure to use cudnn for computational performance
device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

# MODEL_NAME_BLUR2FLOW = "/HPS/HDRVideo2/work/networks/trained_network_blur_to_flow_multscale_1k_slowflow_wocost_50k.pth_epoch12.pth_epoch12.pth"
MODEL_NAME_BLUR2FLOW = "blur2flow.pth"

model_hdrslowmo = blur2flowNet.Network()
model_hdrslowmo.to(device)

def test():
    with torch.no_grad():

        img_hdr = cv2.imread('hdr.png', -1)
        img_blurry = cv2.imread('blurry.png', -1)

        height, width = img_hdr.shape[0:2]

        img_hdr = torch.FloatTensor(
            numpy.array(numpy.asarray(img_hdr).transpose(2, 0, 1).astype(numpy.float32) * (
                    1.0 / 255)))

        img_blurry = torch.FloatTensor(
            numpy.array(numpy.asarray(img_blurry).transpose(2, 0, 1).astype(numpy.float32) * (
                    1.0 / 255)))

        img_hdr = img_hdr.cuda().view(1, 3, int(height), int(width))
        img_blurry = img_blurry.cuda().view(1, 3, int(height), int(width))

        states = torch.load(MODEL_NAME_BLUR2FLOW, map_location='cuda:0')
        model_hdrslowmo.load_state_dict(states['state_model_spatial_flow'])

        flows_curr = model_hdrslowmo(img_blurry, img_hdr)

        blur2flow = torch.nn.functional.interpolate(input=flows_curr[4], size=(int(height), int(width)),
                                                        mode='bilinear',
                                                        align_corners=False) * 4

        blur2flow = blur2flow[0, :, :, :].cpu().detach().numpy().transpose(1, 2, 0)
        blur2flow = flow_vis.flow_to_color(blur2flow, convert_to_bgr=True)

        cv2.imwrite('blur2flow.png', blur2flow)


def main():
    test()

if __name__ == '__main__':
    main()
